@@ -7,6 +7,7 @@
 #include "settings.h"
 #include "globals.h"
 #include "layout.h"
+#include "stack.h"
 #include "utils.h"
 #include "hook.h"
 #include "mouse.h"
@@ -31,16 +32,21 @@
 int g_monitor_float_treshold = 24;
 
 int* g_window_border_width;
+int* g_window_border_inner_width;
 int* g_raise_on_focus;
 int* g_snap_gap;
+int* g_smart_window_surroundings;
 unsigned long g_window_border_active_color;
 unsigned long g_window_border_normal_color;
+unsigned long g_window_border_urgent_color;
+unsigned long g_window_border_inner_color;
 
 GHashTable* g_clients; // container of all clients
 
 
 // atoms from dwm.c
-enum { WMProtocols, WMDelete, WMState, WMLast };        /* default atoms */
+// default atoms
+enum { WMProtocols, WMDelete, WMState, WMTakeFocus, WMLast };
 static Atom g_wmatom[WMLast];
 
 static HSClient* create_client() {
@@ -51,17 +57,25 @@ static HSClient* create_client() {
     hc->urgent = false;
     hc->fullscreen = false;
     hc->pseudotile = false;
+    hc->ewmhrequests = true;
     return hc;
 }
 
 static void fetch_colors() {
     g_window_border_width = &(settings_find("window_border_width")->value.i);
+    g_window_border_inner_width = &(settings_find("window_border_inner_width")->value.i);
+    g_window_gap = &(settings_find("window_gap")->value.i);
     g_snap_gap = &(settings_find("snap_gap")->value.i);
+    g_smart_window_surroundings = &(settings_find("smart_window_surroundings")->value.i);
     g_raise_on_focus = &(settings_find("raise_on_focus")->value.i);
     char* str = settings_find("window_border_normal_color")->value.s;
     g_window_border_normal_color = getcolor(str);
     str = settings_find("window_border_active_color")->value.s;
     g_window_border_active_color = getcolor(str);
+    str = settings_find("window_border_urgent_color")->value.s;
+    g_window_border_urgent_color = getcolor(str);
+    str = settings_find("window_border_inner_color")->value.s;
+    g_window_border_inner_color = getcolor(str);
 }
 
 void clientlist_init() {
@@ -70,9 +84,10 @@ void clientlist_init() {
     g_wmatom[WMProtocols] = XInternAtom(g_display, "WM_PROTOCOLS", False);
     g_wmatom[WMDelete] = XInternAtom(g_display, "WM_DELETE_WINDOW", False);
     g_wmatom[WMState] = XInternAtom(g_display, "WM_STATE", False);
+    g_wmatom[WMTakeFocus] = XInternAtom(g_display, "WM_TAKE_FOCUS", False);
     // init actual client list
     g_clients = g_hash_table_new_full(g_int_hash, g_int_equal,
-                                      NULL, (GDestroyNotify)destroy_client);
+                                      NULL, (GDestroyNotify)client_destroy);
 }
 
 void reset_client_colors() {
@@ -93,21 +108,10 @@ static void client_move_to_floatpos(void* key, void* client_void, void* data) {
     }
 }
 
-static void client_destroy(void* key, void* client_void, void* data) {
-    (void)key;
-    (void)data;
-    HSClient* client = client_void;
-    if (client) {
-        /* free window title */
-        g_string_free(client->title, true);
-    }
-}
-
 void clientlist_destroy() {
     // move all clients to their original floating position
     g_hash_table_foreach(g_clients, client_move_to_floatpos, NULL);
 
-    g_hash_table_foreach(g_clients, client_destroy, NULL);
     g_hash_table_destroy(g_clients);
 }
 
@@ -157,7 +161,7 @@ HSClient* manage_client(Window win) {
 
     if (!changes.manage) {
         client_changes_free_members(&changes);
-        destroy_client(client);
+        client_destroy(client);
         // map it... just to be sure
         XMapWindow(g_display, win);
         return NULL;
@@ -166,14 +170,18 @@ HSClient* manage_client(Window win) {
     // actually manage it
     g_hash_table_insert(g_clients, &(client->window), client);
     ewmh_add_client(client->window);
-    XSetWindowBorderWidth(g_display, win, *g_window_border_width);
     // insert to layout
     if (!client->tag) {
         client->tag = m->tag;
     }
     // get events from window
     XSelectInput(g_display, win, CLIENT_EVENT_MASK);
+    // insert window to the stack
+    client->slice = slice_create_client(client);
+    stack_insert_slice(client->tag->stack, client->slice);
+    // insert windwo to the tag
     frame_insert_window_at_index(client->tag->frame, win, changes.tree_index->str);
+    client_update_wm_hints(client);
     if (changes.focus) {
         // give focus to window if wanted
         // TODO: make this faster!
@@ -186,9 +194,23 @@ HSClient* manage_client(Window win) {
     tag_set_flags_dirty();
     client_set_fullscreen(client, changes.fullscreen);
     ewmh_update_window_state(client);
-    monitor_apply_layout(find_monitor_with_tag(client->tag));
-    client_changes_free_members(&changes);
 
+    HSMonitor* monitor = find_monitor_with_tag(client->tag);
+    if (monitor) {
+        if (monitor != get_current_monitor()
+            && changes.focus && changes.switchtag) {
+            monitor_set_tag(get_current_monitor(), client->tag);
+        }
+        // TODO: monitor_apply_layout() maybe is called twice here if it
+        // already is called by monitor_set_tag()
+        monitor_apply_layout(monitor);
+    } else {
+        if (changes.focus && changes.switchtag) {
+            monitor_set_tag(get_current_monitor(), client->tag);
+        }
+    }
+
+    client_changes_free_members(&changes);
     grab_client_buttons(client, false);
 
     return client;
@@ -201,27 +223,46 @@ void unmanage_client(Window win) {
     }
     // remove from tag
     frame_remove_window(client->tag->frame, win);
-    // and arrange monitor
-    HSMonitor* m = find_monitor_with_tag(client->tag);
-    if (m) monitor_apply_layout(m);
     // ignore events from it
     XSelectInput(g_display, win, 0);
     //XUngrabButton(g_display, AnyButton, AnyModifier, win);
     // permanently remove it
+    HSTag* tag = client->tag;
     g_hash_table_remove(g_clients, &win);
+    // and arrange monitor after the client has been removed from the stack
+    HSMonitor* m = find_monitor_with_tag(tag);
+    tag_update_focus_layer(tag);
+    if (m) monitor_apply_layout(m);
     ewmh_remove_client(win);
     tag_set_flags_dirty();
 }
 
 // destroys a special client
-void destroy_client(HSClient* client) {
+void client_destroy(HSClient* client) {
+    if (client->tag && client->slice) {
+        stack_remove_slice(client->tag->stack, client->slice);
+    }
+    if (client->slice) {
+        slice_destroy(client->slice);
+    }
+    if (client) {
+        /* free window title */
+        g_string_free(client->title, true);
+    }
     g_free(client);
 }
 
 void window_unfocus(Window window) {
-    XSetWindowBorder(g_display, window, g_window_border_normal_color);
+    HSDebug("window_unfocus NORMAL\n");
+    window_update_border(window, g_window_border_normal_color);
     HSClient* c = get_client_from_window(window);
-    if (c) grab_client_buttons(c, false);
+    if (c) {
+        if (c->urgent) {
+            HSDebug("window_unfocus URGENT\n");
+            window_update_border(window, g_window_border_urgent_color);
+        }
+        grab_client_buttons(c, false);
+    }
 }
 
 static Window lastfocus = 0;
@@ -235,37 +276,48 @@ void window_unfocus_last() {
         /* only emit the hook if the focus *really* changes */
         hook_emit_list("focus_changed", "0x0", "", NULL);
         ewmh_update_active_window(None);
+        tag_update_each_focus_layer();
     }
     lastfocus = 0;
 }
 
 void window_focus(Window window) {
-    // unfocus last one
-    window_unfocus(lastfocus);
-    // change window-colors
-    XSetWindowBorder(g_display, window, g_window_border_active_color);
+    HSClient* client = get_client_from_window(window);
+    assert(client != NULL);
     // set keyboardfocus
-    XSetInputFocus(g_display, window, RevertToPointerRoot, CurrentTime);
+    if (!client->neverfocus) {
+        XSetInputFocus(g_display, window, RevertToPointerRoot, CurrentTime);
+    }
+    client_sendevent(client, g_wmatom[WMTakeFocus]);
+
     if (window != lastfocus) {
         /* FIXME: this is a workaround because window_focus always is called
          * twice.  see BUGS for more information
          *
          * only emit the hook if the focus *really* changes */
+        // unfocus last one
+        window_unfocus(lastfocus);
         ewmh_update_active_window(window);
-        HSClient* client = get_client_from_window(window);
+        tag_update_each_focus_layer();
         char* title = client ? client->title->str : "?";
         char winid_str[STRING_BUF_SIZE];
         snprintf(winid_str, STRING_BUF_SIZE, "0x%x", (unsigned int)window);
         hook_emit_list("focus_changed", winid_str, title, NULL);
     }
+
+    // change window-colors
+    HSDebug("window_focus ACTIVE\n");
+    window_update_border(window, g_window_border_active_color);
+
     lastfocus = window;
     /* do some specials for the max layout */
     bool is_max_layout = frame_focused_window(g_cur_frame) == window
                          && g_cur_frame->content.clients.layout == LAYOUT_MAX
                          && get_current_monitor()->tag->floating == false;
     if (*g_raise_on_focus || is_max_layout) {
-        XRaiseWindow(g_display, window);
+        client_raise(client);
     }
+    tag_update_focus_layer(get_current_monitor()->tag);
     grab_client_buttons(get_client_from_window(window), true);
 }
 
@@ -274,7 +326,13 @@ void client_setup_border(HSClient* client, bool focused) {
         g_window_border_normal_color,
         g_window_border_active_color,
     };
-    XSetWindowBorder(g_display, client->window, colors[focused ? 1 : 0]);
+    if (client->urgent) {
+        HSDebug("client_setup_border URGENT\n");
+        window_update_border(client->window, g_window_border_urgent_color);
+    } else {
+        HSDebug("client_setup_border %s\n", (focused ? "ACTIVE" : "NORMAL"));
+        window_update_border(client->window, colors[focused ? 1 : 0]);
+    }
 }
 
 void client_resize_fullscreen(HSClient* client, HSMonitor* m) {
@@ -290,7 +348,12 @@ void client_resize_fullscreen(HSClient* client, HSMonitor* m) {
 
 }
 
-void client_resize(HSClient* client, XRectangle rect) {
+void client_raise(HSClient* client) {
+    assert(client);
+    stack_raise_slide(client->tag->stack, client->slice);
+}
+
+void client_resize(HSClient* client, XRectangle rect, HSFrame* frame) {
     // ensure minimum size
     if (rect.width < WINDOW_MIN_WIDTH) {
         rect.width = WINDOW_MIN_WIDTH;
@@ -305,8 +368,12 @@ void client_resize(HSClient* client, XRectangle rect) {
     Window win = client->window;
     if (client->pseudotile) {
         XRectangle size = client->float_size;
+        // floating sizes don't include window border, tiling sizes do.
+        // so convert the floating size to a tiling size
+        size.width  += *g_window_border_width * 2;
+        size.height += *g_window_border_width * 2;
         // ensure size is not larger than rect-tile
-        size.width = MIN(size.width, rect.width);
+        size.width  = MIN(size.width, rect.width);
         size.height = MIN(size.height, rect.height);
 
         // center it
@@ -315,18 +382,42 @@ void client_resize(HSClient* client, XRectangle rect) {
         rect.width = size.width;
         rect.height = size.height;
     }
+    int border_width = *g_window_border_width;
+    if (*g_smart_window_surroundings && !client->pseudotile
+        && (frame->content.clients.count == 1
+            || frame->content.clients.layout == LAYOUT_MAX)) {
+        border_width = 0;
+    }
     if (RECTANGLE_EQUALS(client->last_size, rect)
-        && client->last_border_width == *g_window_border_width) {
+        && client->last_border_width == border_width) {
         return;
     }
     client->last_size = rect;
-    client->last_border_width = *g_window_border_width;
+    client->last_border_width = border_width;
 
     // apply border width
-    rect.width -= *g_window_border_width * 2;
-    rect.height -= *g_window_border_width * 2;
-    XSetWindowBorderWidth(g_display, win, *g_window_border_width);
+    rect.width -= border_width * 2;
+    rect.height -= border_width * 2;
+    if (!*g_smart_window_surroundings
+        || (frame->content.clients.count != 1
+            && frame->content.clients.layout != LAYOUT_MAX)) {
+        // apply window gap
+        rect.width -= *g_window_gap;
+        rect.height -= *g_window_gap;
+    }
+
+    XSetWindowBorderWidth(g_display, win, border_width);
     XMoveResizeWindow(g_display, win, rect.x, rect.y, rect.width, rect.height);
+    if (*g_window_border_inner_width > 0
+        && *g_window_border_inner_width < *g_window_border_width) {
+        unsigned long current_border_color = get_window_border_color(client);
+        HSDebug("client_resize %s\n",
+                current_border_color == g_window_border_active_color
+                ? "ACTIVE" : "NORMAL");
+        set_window_double_border(g_display, win, *g_window_border_inner_width,
+                                 g_window_border_inner_color,
+                                 current_border_color);
+    }
     //// send new size to client
     //// WHY SHOULD I? -> faster? only one call?
     //XConfigureEvent ce;
@@ -344,12 +435,12 @@ void client_resize(HSClient* client, XRectangle rect) {
     //XSendEvent(g_display, win, False, StructureNotifyMask, (XEvent *)&ce);
 }
 
-void client_resize_tiling(HSClient* client, XRectangle rect) {
+void client_resize_tiling(HSClient* client, XRectangle rect, HSFrame* frame) {
     HSMonitor* m;
     if (client->fullscreen && (m = find_monitor_with_tag(client->tag))) {
         client_resize_fullscreen(client, m);
     } else {
-        client_resize(client, rect);
+        client_resize(client, rect, frame);
     }
 }
 
@@ -385,6 +476,17 @@ void client_resize_floating(HSClient* client, HSMonitor* m) {
     XSetWindowBorderWidth(g_display, client->window, *g_window_border_width);
     XMoveResizeWindow(g_display, client->window,
         rect.x, rect.y, rect.width, rect.height);
+    if (*g_window_border_inner_width > 0
+        && *g_window_border_inner_width < *g_window_border_width) {
+        unsigned long current_border_color = get_window_border_color(client);
+        HSDebug("client_resize %s\n",
+                current_border_color == g_window_border_active_color
+                ? "ACTIVE" : "NORMAL");
+        set_window_double_border(g_display, client->window,
+                                 *g_window_border_inner_width,
+                                 g_window_border_inner_color,
+                                 current_border_color);
+    }
 }
 
 XRectangle client_outer_floating_rect(HSClient* client) {
@@ -459,7 +561,8 @@ void client_update_wm_hints(HSClient* client) {
         return;
     }
 
-    if ((frame_focused_window(g_cur_frame) == client->window)
+    Window focused_window = frame_focused_window(g_cur_frame);
+    if ((focused_window == client->window)
         && wmh->flags & XUrgencyHint) {
         // remove urgency hint if window is focused
         wmh->flags &= ~XUrgencyHint;
@@ -470,10 +573,17 @@ void client_update_wm_hints(HSClient* client) {
             client->urgent = newval;
             char winid_str[STRING_BUF_SIZE];
             snprintf(winid_str, STRING_BUF_SIZE, "0x%lx", client->window);
+            client_setup_border(client, focused_window == client->window);
             hook_emit_list("urgent", client->urgent ? "on":"off", winid_str, NULL);
             tag_set_flags_dirty();
         }
     }
+    if (wmh->flags & InputHint) {
+        client->neverfocus = !wmh->input;
+    } else {
+        client->neverfocus = false;
+    }
+    XFree(wmh);
 }
 
 void client_update_title(HSClient* client) {
@@ -514,10 +624,13 @@ void client_set_fullscreen(HSClient* client, bool state) {
     }
 
     client->fullscreen = state;
+    HSStack* stack = client->tag->stack;
     if (state) {
-        // TODO: do proper stacking layer handling
-        XRaiseWindow(g_display, client->window);
+        stack_slice_add_layer(stack, client->slice, LAYER_FULLSCREEN);
+    } else {
+        stack_slice_remove_layer(stack, client->slice, LAYER_FULLSCREEN);
     }
+    tag_update_focus_layer(client->tag);
     monitor_apply_layout(find_monitor_with_tag(client->tag));
 
     char buf[STRING_BUF_SIZE];
@@ -568,3 +681,93 @@ int client_set_property_command(int argc, char** argv) {
     return 0;
 }
 
+void window_update_border(Window window, unsigned long color) {
+    if (*g_window_border_inner_width > 0
+        && *g_window_border_inner_width < *g_window_border_width) {
+        set_window_double_border(g_display, window,
+                                 *g_window_border_inner_width,
+                                 g_window_border_inner_color, color);
+    } else {
+        XSetWindowBorder(g_display, window, color);
+    }
+}
+
+unsigned long get_window_border_color(HSClient* client) {
+    Window win = client->window;
+    unsigned long current_border_color = (win == frame_focused_window(g_cur_frame)
+                                          ? g_window_border_active_color
+                                          : g_window_border_normal_color);
+    if (client->urgent)
+        current_border_color = g_window_border_urgent_color;
+    return current_border_color;
+}
+
+static bool is_client_urgent(void* key, HSClient* client, void* data) {
+    (void) key;
+    (void) data;
+    return client->urgent;
+}
+
+HSClient* get_urgent_client() {
+    return g_hash_table_find(g_clients, (GHRFunc)is_client_urgent, NULL);
+}
+
+/**
+ * \brief   Resolve a window description to a client or a window id
+ *
+ * \param   str     Describes the window: "" means the focused one, "urgent"
+ *                  resolves to a arbitrary urgent window, "0x..." just
+ *                  resolves to the given window.
+ * \param   ret_client  The client pointer is stored there if ret_client is
+ *                      given and the specified window is managed.
+ * \return          The resolved window id is stored there if the according
+ *                  window has been found
+ */
+Window string_to_client(char* str, HSClient** ret_client) {
+    Window win = 0;
+    if (!strcmp(str, "")) {
+        win = frame_focused_window(g_cur_frame);
+        if (ret_client) {
+            *ret_client = get_client_from_window(win);
+        }
+    } else if (!strcmp(str, "urgent")) {
+        HSClient* client = get_urgent_client();
+        if (client) {
+            win = client->window;
+        }
+        if (ret_client) {
+            *ret_client = client;
+        }
+    } else if (1 == sscanf(str, "0x%lx", (long unsigned int*)&win)) {
+        if (ret_client) {
+            *ret_client = get_client_from_window(win);
+        }
+    }
+    return win;
+}
+
+// mainly from dwm.c
+bool client_sendevent(HSClient *client, Atom proto) {
+    int n;
+    Atom *protocols;
+    bool exists = false;
+    XEvent ev;
+
+    if (XGetWMProtocols(g_display, client->window, &protocols, &n)) {
+        while (!exists && n--)
+            exists = protocols[n] == proto;
+        XFree(protocols);
+    }
+    if (exists) {
+        ev.type = ClientMessage;
+        ev.xclient.window = client->window;
+        ev.xclient.message_type = g_wmatom[WMProtocols];
+        ev.xclient.format = 32;
+        ev.xclient.data.l[0] = proto;
+        ev.xclient.data.l[1] = CurrentTime;
+        XSendEvent(g_display, client->window, False, NoEventMask, &ev);
+    }
+    return exists;
+}
+
+
