@@ -17,6 +17,7 @@
 #include "mouse.h"
 #include "rules.h"
 #include "ewmh.h"
+#include "stack.h"
 // standard
 #include <string.h>
 #include <stdio.h>
@@ -58,6 +59,7 @@ int spawn(int argc, char** argv);
 int wmexec(int argc, char** argv);
 static void remove_zombies(int signal);
 int custom_hook_emit(int argc, char** argv);
+int jumpto_command(int argc, char** argv);
 
 // handler for X-Events
 void buttonpress(XEvent* event);
@@ -83,6 +85,8 @@ CommandBinding g_commands[] = {
     CMD_BIND(             "version",        version),
     CMD_BIND(             "list_commands",  list_commands),
     CMD_BIND(             "list_monitors",  list_monitors),
+    CMD_BIND(             "set_monitors",   set_monitor_rects_command),
+    CMD_BIND(             "disjoin_rects",  disjoin_rects_command),
     CMD_BIND(             "list_keybinds",  key_list_binds),
     CMD_BIND_NO_OUTPUT(   "keybind",        keybind),
     CMD_BIND_NO_OUTPUT(   "keyunbind",      keyunbind),
@@ -91,10 +95,13 @@ CommandBinding g_commands[] = {
     CMD_BIND_NO_OUTPUT(   "spawn",          spawn),
     CMD_BIND_NO_OUTPUT(   "wmexec",         wmexec),
     CMD_BIND_NO_OUTPUT(   "emit_hook",      custom_hook_emit),
+    CMD_BIND_NO_OUTPUT(   "bring",          frame_current_bring),
+    CMD_BIND_NO_OUTPUT(   "focus_nth",      frame_current_set_selection),
     CMD_BIND_NO_OUTPUT(   "cycle",          frame_current_cycle_selection),
     CMD_BIND_NO_OUTPUT(   "cycle_all",      cycle_all_command),
     CMD_BIND_NO_OUTPUT(   "cycle_layout",   frame_current_cycle_client_layout),
     CMD_BIND_NO_OUTPUT(   "close",          window_close_current),
+    CMD_BIND_NO_OUTPUT(   "close_or_remove",close_or_remove_command),
     CMD_BIND_NO_OUTPUT(   "split",          frame_split_command),
     CMD_BIND_NO_OUTPUT(   "resize",         frame_change_fraction_command),
     CMD_BIND_NO_OUTPUT(   "focus",          frame_focus_command),
@@ -109,6 +116,7 @@ CommandBinding g_commands[] = {
     CMD_BIND_NO_OUTPUT(   "add",            tag_add_command),
     CMD_BIND_NO_OUTPUT(   "use",            monitor_set_tag_command),
     CMD_BIND_NO_OUTPUT(   "use_index",      monitor_set_tag_by_index_command),
+    CMD_BIND_NO_OUTPUT(   "jumpto",         jumpto_command),
     CMD_BIND(             "floating",       tag_set_floating_command),
     CMD_BIND_NO_OUTPUT(   "fullscreen",     client_set_property_command),
     CMD_BIND_NO_OUTPUT(   "pseudotile",     client_set_property_command),
@@ -116,8 +124,10 @@ CommandBinding g_commands[] = {
     CMD_BIND_NO_OUTPUT(   "merge_tag",      tag_remove_command),
     CMD_BIND_NO_OUTPUT(   "rename",         tag_rename_command),
     CMD_BIND_NO_OUTPUT(   "move",           tag_move_window_command),
+    CMD_BIND_NO_OUTPUT(   "rotate",         layout_rotate_command),
     CMD_BIND_NO_OUTPUT(   "move_index",     tag_move_window_by_index_command),
     CMD_BIND_NO_OUTPUT(   "add_monitor",    add_monitor_command),
+    CMD_BIND_NO_OUTPUT(   "raise_monitor",  monitor_raise_command),
     CMD_BIND_NO_OUTPUT(   "remove_monitor", remove_monitor_command),
     CMD_BIND_NO_OUTPUT(   "move_monitor",   move_monitor_command),
     CMD_BIND(             "monitor_rect",   monitor_rect_command),
@@ -126,12 +136,14 @@ CommandBinding g_commands[] = {
     CMD_BIND_NO_OUTPUT(   "rule",           rule_add_command),
     CMD_BIND_NO_OUTPUT(   "unrule",         rule_remove_command),
     CMD_BIND(             "layout",         print_layout_command),
+    CMD_BIND(             "stack",          print_stack_command),
     CMD_BIND(             "dump",           print_layout_command),
     CMD_BIND(             "load",           load_command),
     CMD_BIND(             "complete",       complete_command),
     CMD_BIND_NO_OUTPUT(   "lock",           monitors_lock_command),
     CMD_BIND_NO_OUTPUT(   "unlock",         monitors_unlock_command),
     CMD_BIND_NO_OUTPUT(   "set_layout",     frame_current_set_client_layout),
+    CMD_BIND_NO_OUTPUT(   "detect_monitors",detect_monitors_command),
     {{ NULL }}
 };
 
@@ -163,7 +175,7 @@ int print_layout_command(int argc, char** argv, GString** result) {
     }
     // if no tag was found
     if (!tag) {
-        HSMonitor* m = &g_array_index(g_monitors, HSMonitor, g_cur_monitor);
+        HSMonitor* m = get_current_monitor();
         tag = m->tag;
     }
     assert(tag != NULL);
@@ -188,7 +200,7 @@ int load_command(int argc, char** argv, GString** result) {
     }
     // if no tag was found
     if (!tag) {
-        HSMonitor* m = &g_array_index(g_monitors, HSMonitor, g_cur_monitor);
+        HSMonitor* m = get_current_monitor();
         tag = m->tag;
     }
     assert(tag != NULL);
@@ -223,9 +235,9 @@ int print_tag_status_command(int argc, char** argv, GString** result) {
     if (argc >= 2) {
         monitor_index = atoi(argv[1]);
     }
-    monitor_index = CLAMP(monitor_index, 0, g_monitors->len);
+    monitor_index = CLAMP(monitor_index, 0, monitor_count());
     tag_update_flags();
-    HSMonitor* monitor = &g_array_index(g_monitors, HSMonitor, monitor_index);
+    HSMonitor* monitor = monitor_with_index(monitor_index);
     *result = g_string_append_c(*result, '\t');
     for (int i = 0; i < g_tags->len; i++) {
         HSTag* tag = g_array_index(g_tags, HSTag*, i);
@@ -234,11 +246,17 @@ int print_tag_status_command(int argc, char** argv, GString** result) {
         if (tag->flags & TAG_FLAG_USED) {
             c = ':';
         }
-        if (tag == monitor->tag) {
+        HSMonitor *tag_monitor = find_monitor_with_tag(tag);
+        if (tag_monitor == monitor) {
             c = '+';
             if (monitor_index == g_cur_monitor) {
                 c = '#';
             }
+        } else if (tag_monitor) {
+            c = '-';
+            if (get_current_monitor() == tag_monitor) {
+                c = '%';
+            }
         }
         if (tag->flags & TAG_FLAG_URGENT) {
             c = '!';
@@ -309,21 +327,26 @@ int wmexec(int argc, char** argv) {
 
 int raise_command(int argc, char** argv) {
     Window win;
-    if (argc > 1) {
-        if (1 != sscanf(argv[1], "0x%lx", (long unsigned int*)&win)) {
-            // a invalid winid was given
-            return 1;
-        }
+    HSClient* client = NULL;
+    win = string_to_client((argc > 1) ? argv[1] : "", &client);
+if (client) {
+        client_raise(client);
     } else {
-        win = frame_focused_window(g_cur_frame);
-        if (!win) {
-            return 0;
-        }
+        XRaiseWindow(g_display, win);
     }
-    XRaiseWindow(g_display, win);
     return 0;
 }
 
+int jumpto_command(int argc, char** argv) {
+    HSClient* client = NULL;
+    string_to_client((argc > 1) ? argv[1] : "", &client);
+    if (client) {
+        focus_window(client->window, true, true);
+    }
+    return 0;
+}
+
+
 // handle x-events:
 
 void event_on_configure(XEvent event) {
@@ -335,6 +358,10 @@ void event_on_configure(XEvent event) {
     ce.event = cre->window;
     ce.window = cre->window;
     if (client) {
+        cre->width += 2*cre->border_width - 2*client->last_border_width;
+        cre->height += 2*cre->border_width - 2*client->last_border_width;
+        client->float_size.width = cre->width;
+        client->float_size.height = cre->height;
         ce.x = client->last_size.x;
         ce.y = client->last_size.y;
         ce.width = client->last_size.width;
@@ -342,8 +369,13 @@ void event_on_configure(XEvent event) {
         ce.override_redirect = False;
         ce.border_width = cre->border_width;
         ce.above = cre->above;
+        if (client->tag->floating || client->pseudotile) {
+            monitor_apply_layout(find_monitor_with_tag(client->tag));
+        } else {
         // FIXME: why send event and not XConfigureWindow or XMoveResizeWindow??
-        XSendEvent(g_display, cre->window, False, StructureNotifyMask, (XEvent*)&ce);
+            XSendEvent(g_display, cre->window, False, StructureNotifyMask,
+                       (XEvent*)&ce);
+        }
     } else {
         // if client not known.. then allow configure.
         // its probably a nice conky or dzen2 bar :)
@@ -366,6 +398,8 @@ void event_on_configure(XEvent event) {
  * default error handler, which may call exit.  */
 int xerror(Display *dpy, XErrorEvent *ee) {
     if(ee->error_code == BadWindow
+    || ee->error_code == BadGC
+    || ee->error_code == BadPixmap
     || (ee->request_code == X_SetInputFocus && ee->error_code == BadMatch)
     || (ee->request_code == X_PolyText8 && ee->error_code == BadDrawable)
     || (ee->request_code == X_PolyFillRectangle && ee->error_code == BadDrawable)
@@ -373,8 +407,9 @@ int xerror(Display *dpy, XErrorEvent *ee) {
     || (ee->request_code == X_ConfigureWindow && ee->error_code == BadMatch)
     || (ee->request_code == X_GrabButton && ee->error_code == BadAccess)
     || (ee->request_code == X_GrabKey && ee->error_code == BadAccess)
-    || (ee->request_code == X_CopyArea && ee->error_code == BadDrawable))
+    || (ee->request_code == X_CopyArea && ee->error_code == BadDrawable)) {
         return 0;
+    }
     fprintf(stderr, "herbstluftwm: fatal error: request code=%d, error code=%d\n",
             ee->request_code, ee->error_code);
     if (ee->error_code == BadDrawable) {
@@ -415,20 +450,26 @@ void checkotherwm(void) {
 // scan for windows and add them to the list of managed clients
 // from dwm.c
 void scan(void) {
-    unsigned int i, num;
-    Window d1, d2, *wins = NULL;
+    unsigned int num;
+    Window d1, d2, *cl, *wins = NULL;
+    unsigned long cl_count;
     XWindowAttributes wa;
 
-    if(XQueryTree(g_display, g_root, &d1, &d2, &wins, &num)) {
-        for(i = 0; i < num; i++) {
+    ewmh_get_original_client_list(&cl, &cl_count);
+    if (XQueryTree(g_display, g_root, &d1, &d2, &wins, &num)) {
+        for (int i = 0; i < num; i++) {
             if(!XGetWindowAttributes(g_display, wins[i], &wa)
             || wa.override_redirect || XGetTransientForHint(g_display, wins[i], &d1))
                 continue;
             // only manage mapped windows.. no strange wins like:
             //      luakit/dbus/(ncurses-)vim
+            // but manage it if it was in the ewmh property _NET_CLIENT_LIST by
+            // the previous window manager
             // TODO: what would dwm do?
-            if (is_window_mapped(g_display, wins[i])) {
+            if (is_window_mapped(g_display, wins[i])
+                || 0 <= array_find(cl, cl_count, sizeof(Window), wins+i)) {
                 manage_client(wins[i]);
+                XMapWindow(g_display, wins[i]);
             }
         }
         if(wins)
@@ -437,7 +478,7 @@ void scan(void) {
 }
 
 void execute_autostart_file() {
-    GString* path;
+    GString* path = NULL;
     if (g_autostart_path) {
         path = g_string_new(g_autostart_path);
     } else {
@@ -459,11 +500,21 @@ void execute_autostart_file() {
         path = g_string_append_c(path, G_DIR_SEPARATOR);
         path = g_string_append(path, HERBSTLUFT_AUTOSTART);
     }
-    char* argv[] = {
-        "...", // command name... but it doesnot matter
-        path->str
-    };
-    spawn(LENGTH(argv), argv);
+    if (0 == fork()) {
+        if (g_display) {
+            close(ConnectionNumber(g_display));
+        }
+        setsid();
+        execl(path->str, path->str, NULL);
+
+        char* global_autostart = HERBSTLUFT_GLOBAL_AUTOSTART;
+        HSDebug("Can not execute %s, falling back to %s\n", path->str, global_autostart);
+        execl(global_autostart, global_autostart, NULL);
+
+        fprintf(stderr, "herbstluftwm: execvp \"%s\"", global_autostart);
+        perror(" failed");
+        exit(EXIT_FAILURE);
+    }
     g_string_free(path, true);
 }
 
@@ -555,8 +606,11 @@ static struct {
     { ipc_init,         ipc_destroy         },
     { key_init,         key_destroy         },
     { settings_init,    settings_destroy    },
-    { clientlist_init,  clientlist_destroy  },
+    { stacklist_init,   stacklist_destroy   },
     { layout_init,      layout_destroy      },
+    { tag_init,         tag_destroy         },
+    { clientlist_init,  clientlist_destroy  },
+    { monitor_init,     monitor_destroy     },
     { ewmh_init,        ewmh_destroy        },
     { mouse_init,       mouse_destroy       },
     { hook_init,        hook_destroy        },
@@ -575,7 +629,10 @@ void buttonpress(XEvent* event) {
     } else {
         focus_window(be->window, false, true);
         if (*g_raise_on_click) {
-            XRaiseWindow(g_display, be->window);
+            HSClient* client = get_client_from_window(be->window);
+            if (client) {
+                client_raise(client);
+            }
         }
     }
     XAllowEvents(g_display, ReplayPointer, be->time);
@@ -597,6 +654,11 @@ void configurerequest(XEvent* event) {
     event_on_configure(*event);
 }
 void configurenotify(XEvent* event) {
+    if (event->xconfigure.window == g_root &&
+        settings_find("auto_detect_monitors")->value.i) {
+        char* args[] = { "detect_monitors" };
+        detect_monitors_command(LENGTH(args), args);
+    }
     // HSDebug("name is: ConfigureNotify\n");
 }
 void destroynotify(XEvent* event) {
@@ -763,11 +825,13 @@ int main(int argc, char* argv[]) {
     if (g_exec_before_quit) {
         if (g_exec_args) {
             // do actual exec
+            HSDebug("==> Doing wmexec to %s\n", g_exec_args[0]);
             execvp(g_exec_args[0], g_exec_args);
             fprintf(stderr, "herbstluftwm: execvp \"%s\"", g_exec_args[0]);
             perror(" failed");
         }
         // on failure or if no other wm given, then fall back
+        HSDebug("==> Doing wmexec to %s\n", argv[0]);
         execvp(argv[0], argv);
         fprintf(stderr, "herbstluftwm: execvp \"%s\"", argv[1]);
         perror(" failed");
