@@ -11,7 +11,6 @@
 #include "ipc-protocol.h"
 
 #include <glib.h>
-#include <time.h>
 #include <string.h>
 #include <stdio.h>
 #include <sys/types.h>
@@ -31,52 +30,60 @@ typedef struct {
 } HSConsequenceType;
 
 /// DECLARATIONS ///
-
 static int find_condition_type(char* name);
 static int find_consequence_type(char* name);
 static bool condition_string(HSCondition* rule, char* string);
-static bool condition_class(HSCondition* rule, HSClient* client);
-static bool condition_instance(HSCondition* rule, HSClient* client);
-static bool condition_title(HSCondition* rule, HSClient* client);
-static bool condition_pid(HSCondition* rule, HSClient* client);
-static bool condition_maxage(HSCondition* rule, HSClient* client);
-static bool condition_windowtype(HSCondition* rule, HSClient* client);
-static bool condition_windowrole(HSCondition* rule, HSClient* client);
-static void consequence_tag(HSConsequence* cons, HSClient* client,
-                            HSClientChanges* changes);
-static void consequence_focus(HSConsequence* cons, HSClient* client,
-                              HSClientChanges* changes);
-static void consequence_manage(HSConsequence* cons, HSClient* client,
-                              HSClientChanges* changes);
-static void consequence_index(HSConsequence* cons, HSClient* client,
-                              HSClientChanges* changes);
-static void consequence_pseudotile(HSConsequence* cons, HSClient* client,
-                                   HSClientChanges* changes);
-static void consequence_fullscreen(HSConsequence* cons, HSClient* client,
-                                   HSClientChanges* changes);
+
+/// CONDITIONS ///
+#define DECLARE_CONDITION(NAME)                         \
+    static bool NAME(HSCondition* rule, HSClient* client)
+
+DECLARE_CONDITION(condition_class);
+DECLARE_CONDITION(condition_instance);
+DECLARE_CONDITION(condition_title);
+DECLARE_CONDITION(condition_pid);
+DECLARE_CONDITION(condition_maxage);
+DECLARE_CONDITION(condition_windowtype);
+DECLARE_CONDITION(condition_windowrole);
+
+/// CONSEQUENCES ///
+#define DECLARE_CONSEQUENCE(NAME)                           \
+static void NAME(HSConsequence* cons, HSClient* client,     \
+                     HSClientChanges* changes)
+
+DECLARE_CONSEQUENCE(consequence_tag);
+DECLARE_CONSEQUENCE(consequence_focus);
+DECLARE_CONSEQUENCE(consequence_manage);
+DECLARE_CONSEQUENCE(consequence_index);
+DECLARE_CONSEQUENCE(consequence_pseudotile);
+DECLARE_CONSEQUENCE(consequence_fullscreen);
+DECLARE_CONSEQUENCE(consequence_switchtag);
+DECLARE_CONSEQUENCE(consequence_ewmhrequests);
 
 /// GLOBALS ///
 
 static HSConditionType g_condition_types[] = {
-    {   "class",    condition_class },
-    {   "instance", condition_instance },
-    {   "title",    condition_title },
-    {   "pid",      condition_pid },
-    {   "maxage",   condition_maxage },
-    {   "windowtype", condition_windowtype },
-    {   "windowrole", condition_windowrole },
+    { "class",          condition_class             },
+    { "instance",       condition_instance          },
+    { "title",          condition_title             },
+    { "pid",            condition_pid               },
+    { "maxage",         condition_maxage            },
+    { "windowtype",     condition_windowtype        },
+    { "windowrole",     condition_windowrole        },
 };
 
 int     g_maxage_type; // index of "maxage"
 time_t  g_current_rule_birth_time; // data from rules_apply() to condition_maxage()
 
 static HSConsequenceType g_consequence_types[] = {
-    {   "tag",          consequence_tag },
-    {   "index",        consequence_index },
-    {   "focus",        consequence_focus },
-    {   "manage",       consequence_manage },
-    {   "pseudotile",   consequence_pseudotile },
-    {   "fullscreen",   consequence_fullscreen },
+    { "tag",            consequence_tag             },
+    { "index",          consequence_index           },
+    { "focus",          consequence_focus           },
+    { "switchtag",      consequence_switchtag       },
+    { "manage",         consequence_manage          },
+    { "pseudotile",     consequence_pseudotile      },
+    { "fullscreen",     consequence_fullscreen      },
+    { "ewmhrequests",   consequence_ewmhrequests    },
 };
 
 GQueue g_rules = G_QUEUE_INIT; // a list of HSRule* elements
@@ -220,9 +227,7 @@ void consequence_destroy(HSConsequence* cons) {
 HSRule* rule_create() {
     HSRule* rule = g_new0(HSRule, 1);
     rule->once = false;
-    struct timespec t;
-    clock_gettime(CLOCK_MONOTONIC, &t);
-    rule->birth_time = t.tv_sec;
+    rule->birth_time = get_monotonic_timestamp();
     return rule;
 }
 
@@ -371,6 +376,7 @@ void client_changes_init(HSClientChanges* changes, HSClient* client) {
     memset(changes, 0, sizeof(HSClientChanges));
     changes->tree_index = g_string_new("");
     changes->focus = false;
+    changes->switchtag = false;
     changes->manage = true;
     changes->fullscreen = ewmh_is_fullscreen_set(client->window);
 }
@@ -510,9 +516,7 @@ bool condition_pid(HSCondition* rule, HSClient* client) {
 }
 
 bool condition_maxage(HSCondition* rule, HSClient* client) {
-    struct timespec cur;
-    clock_gettime(CLOCK_MONOTONIC, &cur);
-    time_t diff = cur.tv_sec - g_current_rule_birth_time;
+    time_t diff = get_monotonic_timestamp() - g_current_rule_birth_time;
     return (rule->value.integer >= diff);
 }
 
@@ -612,3 +616,16 @@ void consequence_fullscreen(HSConsequence* cons, HSClient* client,
     changes->fullscreen = string_to_bool(cons->value.str, changes->fullscreen);
 }
 
+void consequence_switchtag(HSConsequence* cons, HSClient* client,
+                           HSClientChanges* changes) {
+    changes->switchtag = string_to_bool(cons->value.str, changes->switchtag);
+}
+
+void consequence_ewmhrequests(HSConsequence* cons, HSClient* client,
+                              HSClientChanges* changes) {
+    // this is only a flag that is unused during initialization (during
+    // manage()) and so can be directly changed in the client
+    client->ewmhrequests = string_to_bool(cons->value.str, client->ewmhrequests);
+}
+
+
