@@ -30,6 +30,19 @@ GString* window_class_to_g_string(Display* dpy, Window window);
 GString* window_instance_to_g_string(Display* dpy, Window window);
 int window_pid(Display* dpy, Window window);
 
+typedef void* HSTree;
+struct HSTreeInterface;
+typedef struct HSTreeInterface {
+    struct HSTreeInterface  (*nth_child)(HSTree root, size_t idx);
+    size_t                  (*child_count)(HSTree root);
+    void                    (*append_caption)(HSTree root, GString** output);
+    HSTree                  data;
+    void                    (*destructor)(HSTree data); /* how to free the data tree */
+} HSTreeInterface;
+
+void tree_print_to(HSTreeInterface* intface, GString** output);
+
+
 bool is_herbstluft_window(Display* dpy, Window window);
 
 bool is_window_mapable(Display* dpy, Window window);
@@ -41,6 +54,8 @@ bool string_to_bool(char* string, bool oldvalue);
 
 char* strlasttoken(char* str, char* delim);
 
+time_t get_monotonic_timestamp();
+
 // duplicates an argument-vector
 char** argv_duplicate(int argc, char** argv);
 // frees all entrys in argument-vector and then the vector itself
@@ -52,6 +67,7 @@ void g_queue_remove_element(GQueue* queue, GList* elem);
 
 // find an element in an array buf with elems elements of size size.
 int array_find(void* buf, size_t elems, size_t size, void* needle);
+void array_reverse(void* buf, size_t elems, size_t size);
 
 
 // utils for tables
@@ -61,6 +77,10 @@ bool memberequals_int(void* pmember, void* needle);
 
 void* table_find(void* start, size_t elem_size, size_t count,
                  size_t member_offset, MemberEquals equals, void* needle);
+
+void set_window_double_border(Display *dpy, Window win, int ibw,
+                              unsigned long inner_color, unsigned long outer_color);
+
 #define STATIC_TABLE_FIND(TYPE, TABLE, MEMBER, EQUALS, NEEDLE)  \
     ((TYPE*) table_find((TABLE),                                \
                         sizeof(TABLE[0]),                       \
